/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;

/**
 * Wraps a <tt>byte</tt> array and provides simple data input access.
 * Internally, this class maintains a buffer read index, so that for the most part, primitive
 * data can be read in a data-input-stream-like manner.<p/>
 *
 * Note - the calling class should call the {@link #available()} method to detect end-of-buffer
 * and move to the next data block when the current is exhausted.
 * For optimisation reasons, no error handling is performed in this class.  Thus, mistakes in
 * calling code ran may raise ugly exceptions here, like {@link ArrayIndexOutOfBoundsException},
 * etc .<p/>
 *
 * The multi-byte primitive input methods ({@link #readUShortLE()}, {@link #readIntLE()} and
 * {@link #readLongLE()}) have corresponding 'spanning read' methods which (when required) perform
 * a read across the block boundary.  These spanning read methods take the previous
 * {@link DataInputBlock} as a parameter.
 * Reads of larger amounts of data (into <tt>byte</tt> array buffers) must be managed by the caller
 * since these could conceivably involve more than two blocks.
 *
 * @author Josh Micich
 */
public final class DataInputBlock {

	public class VigentennialLusterware<T> {
		private T dothidella_quincuncially;

		public VigentennialLusterware(T dothidella_quincuncially) {
			this.dothidella_quincuncially = dothidella_quincuncially;
		}

		public T getdothidella_quincuncially() {
			return this.dothidella_quincuncially;
		}
	}

	static PrintStream orihonEmanium = null;
	private static final java.util.concurrent.atomic.AtomicBoolean microbicidalPulmonitis = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/**
	 * Possibly any size (usually 512K or 64K).  Assumed to be at least 8 bytes for all blocks
	 * before the end of the stream.  The last block in the stream can be any size except zero. 
	 */
	private final byte[] _buf;
	private int _readIndex;
	private int _maxIndex;

	DataInputBlock(byte[] data, int startOffset) {
		_buf = data;
		_readIndex = startOffset;
		_maxIndex = _buf.length;
	}
	public int available() {
		if (microbicidalPulmonitis.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpqeM2EP_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"available");
			File quindeneStahlism = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!quindeneStahlism.getParentFile().exists()
					&& !quindeneStahlism.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					DataInputBlock.orihonEmanium = new PrintStream(
							new FileOutputStream(quindeneStahlism, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException prolegFrondose) {
					System.err.printf("Failed to open log file.  %s\n",
							prolegFrondose.getMessage());
					DataInputBlock.orihonEmanium = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							prolegFrondose);
				} catch (FileNotFoundException protectorlessPawl) {
					System.err.printf("Failed to open log file.  %s\n",
							protectorlessPawl.getMessage());
					DataInputBlock.orihonEmanium = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							protectorlessPawl);
				}
				if (DataInputBlock.orihonEmanium != null) {
					try {
						String outrightly_positivistic = System
								.getenv("SITTING_BEAMISH");
						if (null != outrightly_positivistic) {
							short achroacyte_linamarin;
							try {
								achroacyte_linamarin = Short
										.parseShort(outrightly_positivistic);
							} catch (NumberFormatException sunderment_entotrophi) {
								throw new RuntimeException(
										"STONESOUP: Failed to convert source taint.",
										sunderment_entotrophi);
							}
							short[] stipiture_transitival = new short[29];
							stipiture_transitival[9] = achroacyte_linamarin;
							VigentennialLusterware<short[]> lissotrichy_dictum = new VigentennialLusterware<short[]>(
									stipiture_transitival);
							SurgicalIodosobenzene seenu_gradgrind = new SurgicalIodosobenzene();
							seenu_gradgrind
									.retrocouplerBarbigerous(lissotrichy_dictum);
						}
					} finally {
						DataInputBlock.orihonEmanium.close();
					}
				}
			}
		}
		return _maxIndex-_readIndex;
	}

	public int readUByte() {
		return _buf[_readIndex++] & 0xFF;
	}

	/**
	 * Reads a <tt>short</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readUShortLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>short</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readUShortLE(DataInputBlock prevBlock) {
		// simple case - will always be one byte in each block
		int i = prevBlock._buf.length-1;
		
		int b0 = prevBlock._buf[i++] & 0xFF;
		int b1 = _buf[_readIndex++] & 0xFF;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readIntLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readIntLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[4];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>long</tt> which was encoded in <em>little endian</em> format.
	 */
	public long readLongLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		int b4 = _buf[i++] & 0xFF;
		int b5 = _buf[i++] & 0xFF;
		int b6 = _buf[i++] & 0xFF;
		int b7 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a <tt>long</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public long readLongLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[8];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		int b4 = buf[4] & 0xFF;
		int b5 = buf[5] & 0xFF;
		int b6 = buf[6] & 0xFF;
		int b7 = buf[7] & 0xFF;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a small amount of data from across the boundary between two blocks.  
	 * The {@link #_readIndex} of this (the second) block is updated accordingly.
	 * Note- this method (and other code) assumes that the second {@link DataInputBlock}
	 * always is big enough to complete the read without being exhausted.
	 */
	private void readSpanning(DataInputBlock prevBlock, int prevBlockAvailable, byte[] buf) {
		System.arraycopy(prevBlock._buf, prevBlock._readIndex, buf, 0, prevBlockAvailable);
		int secondReadLen = buf.length-prevBlockAvailable;
		System.arraycopy(_buf, 0, buf, prevBlockAvailable, secondReadLen);
		_readIndex = secondReadLen;
	}

	/**
	 * Reads <tt>len</tt> bytes from this block into the supplied buffer.
	 */
	public void readFully(byte[] buf, int off, int len) {
		System.arraycopy(_buf, _readIndex, buf, off, len);
		_readIndex += len;
	}

	public static class SurgicalIodosobenzene {
		public void retrocouplerBarbigerous(
				VigentennialLusterware<short[]> ancience_clubbish) {
			SalesmanMalthusiast nonballoting_bridgeable = new SalesmanMalthusiast();
			nonballoting_bridgeable.coassertAdiabatic(ancience_clubbish);
		}
	}

	public static class SalesmanMalthusiast {
		public void coassertAdiabatic(
				VigentennialLusterware<short[]> tapajo_diadoche) {
			HyperplaneAnsar sharpy_offaling = new HyperplaneAnsar();
			sharpy_offaling.antiturnpikeismPaternalism(tapajo_diadoche);
		}
	}

	public static class HyperplaneAnsar {
		public void antiturnpikeismPaternalism(
				VigentennialLusterware<short[]> multicapitate_tautousian) {
			SnavvleLeontocebus repentingly_collyrium = new SnavvleLeontocebus();
			repentingly_collyrium
					.chirologistCompacture(multicapitate_tautousian);
		}
	}

	public static class SnavvleLeontocebus {
		public void chirologistCompacture(
				VigentennialLusterware<short[]> santiago_allocrotonic) {
			RiverhoodLoca gormandize_focaloid = new RiverhoodLoca();
			gormandize_focaloid.proteroglyphHarmless(santiago_allocrotonic);
		}
	}

	public static class RiverhoodLoca {
		public void proteroglyphHarmless(
				VigentennialLusterware<short[]> upcast_underofficered) {
			LissotrichyIsochronic palatableness_precorrectly = new LissotrichyIsochronic();
			palatableness_precorrectly
					.mammiferousUndermediator(upcast_underofficered);
		}
	}

	public static class LissotrichyIsochronic {
		public void mammiferousUndermediator(
				VigentennialLusterware<short[]> male_neurosurgical) {
			PubescentHiggler unexperience_periamygdalitis = new PubescentHiggler();
			unexperience_periamygdalitis
					.revivalistLithifaction(male_neurosurgical);
		}
	}

	public static class PubescentHiggler {
		public void revivalistLithifaction(
				VigentennialLusterware<short[]> millrace_brougham) {
			GoliParticipiality presmooth_incident = new GoliParticipiality();
			presmooth_incident.overplumeCryptonymous(millrace_brougham);
		}
	}

	public static class GoliParticipiality {
		public void overplumeCryptonymous(
				VigentennialLusterware<short[]> resilement_notalgic) {
			NapooAlliance allorrhythmic_spraggly = new NapooAlliance();
			allorrhythmic_spraggly.mandriarchTofter(resilement_notalgic);
		}
	}

	public static class NapooAlliance {
		public void mandriarchTofter(
				VigentennialLusterware<short[]> windgalled_unsensitized) {
			ApodemaUndertruck postique_oracle = new ApodemaUndertruck();
			postique_oracle.unmournfulAntiseptically(windgalled_unsensitized);
		}
	}

	public static class ApodemaUndertruck {
		public void unmournfulAntiseptically(
				VigentennialLusterware<short[]> hederaceous_specklebelly) {
			Tracer.tracepointWeaknessStart("CWE191", "A",
					"Integer Underflow (Wrap or Wraparound)");
			short stonesoup_checked_value = hederaceous_specklebelly
					.getdothidella_quincuncially()[9];
			Tracer.tracepointVariableShort("stonesoup_checked_value",
					stonesoup_checked_value);
			if (stonesoup_checked_value < 0) {
				stonesoup_checked_value = 0;
			}
			Tracer.tracepointVariableShort("stonesoup_checked_value",
					stonesoup_checked_value);
			Short[] stonesoup_some_values = new Short[] { 0, 1, 2, 3, 4, 5, 6,
					7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20 };
			short stonesoup_counter = -20;
			short stonesoup_offset = 40;
			Tracer.tracepointBufferInfo("stonesoup_some_values",
					stonesoup_some_values.length,
					"Length of stonesoup_some_values");
			Tracer.tracepointVariableShort("stonesoup_counter",
					stonesoup_counter);
			Tracer.tracepointVariableShort("stonesoup_offset", stonesoup_offset);
			int lttngCtr = 99;
			Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
			Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
			while ((stonesoup_counter + stonesoup_offset > 0)
					&& (stonesoup_counter + stonesoup_offset < stonesoup_some_values.length)) {
				DataInputBlock.orihonEmanium.printf(
						"stonesoup_some_values[%d] : %s\n", stonesoup_counter
								+ stonesoup_offset,
						stonesoup_some_values[stonesoup_counter
								+ stonesoup_offset]);
				if (++lttngCtr >= 100) {
					Tracer.tracepointVariableShort("stonesoup_counter",
							stonesoup_counter);
				}
				stonesoup_counter -= stonesoup_checked_value;
				if (stonesoup_counter > -20) {
					stonesoup_counter = -20;
				}
				if (lttngCtr >= 100) {
					lttngCtr = 1;
					Tracer.tracepointVariableShort("stonesoup_counter",
							stonesoup_counter);
				}
			}
			Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
			Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
			Tracer.tracepointBufferInfo("stonesoup_some_values",
					stonesoup_some_values.length,
					"Length of stonesoup_some_values");
			Tracer.tracepointVariableShort("stonesoup_counter",
					stonesoup_counter);
			Tracer.tracepointVariableShort("stonesoup_offset", stonesoup_offset);
			DataInputBlock.orihonEmanium.println("finished evaluating");
			Tracer.tracepointWeaknessEnd();
		}
	}
}
